import React, { useEffect, useMemo, useRef, useState } from 'react';
import { Table, AugmentedNodeAttributes } from './components/Table';
import { SchemaAttribute, SchemaNode, SchemaAttributeTypes } from '../../../schema';
import { VisualizationPropTypes, VISComponentType } from '../../common';
import { Input } from '@graphpolaris/shared/lib/components/inputs';
import { GraphMetadata } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer } from '@graphpolaris/shared/lib/vis/components/config';

export type SchemaAttributeCustom = {
  name: string;
  type: SchemaAttributeTypes;
};

export type TableProps = {
  showBarplot: boolean;
  itemsPerPage: number;
  displayAttributes: string[];
  displayEntity: string;
};

const configuration: TableProps = {
  itemsPerPage: 10,
  showBarplot: true,
  displayAttributes: [],
  displayEntity: '',
};

export const TableVis = ({ data, schema, configuration, updateSettings, graphMetadata }: VisualizationPropTypes) => {
  const ref = useRef<HTMLDivElement>(null);
  const prevDisplayAttributes = useRef<string[]>(configuration.displayAttributes);
  const [savedDisplayAttributes, setSavedDisplayAttributes] = useState<string[]>(configuration.displayAttributes);

  /*
  useEffect(() => {
    setSavedDisplayAttributes(configuration.displayAttributes);
  }, [configuration.displayAttributes]);
  */
  /*
  useEffect(() => {
    if (!graphMetadata.nodes.labels.includes(configuration.displayEntity)) {
      updateSettings({ displayEntity: graphMetadata.nodes.labels[0] });
    }

  }, [graphMetadata.nodes.labels, data, configuration]);
  */

  useEffect(() => {
    const filteredDisplayAttributes = prevDisplayAttributes.current.filter((attr) => configuration.displayAttributes.includes(attr));
    setSavedDisplayAttributes(filteredDisplayAttributes);
  }, [configuration.displayAttributes]);

  const attributesArray = useMemo<AugmentedNodeAttributes[]>(() => {
    const filteredDisplayAttributes = prevDisplayAttributes.current.filter((attr) => configuration.displayAttributes.includes(attr));

    return data.nodes
      .filter((node) => {
        // some dataset do not have label field
        let labelNode = '';
        if (node.label !== undefined) {
          labelNode = node.label;
        } else {
          const idParts = node._id.split('/');
          labelNode = idParts[0];
        }
        return labelNode === configuration.displayEntity;
      })
      .map((node) => {
        // get attributes filtered and sorted
        const filteredAttributes = Object.fromEntries(
          Object.entries(node.attributes)
            .filter(([attr]) => filteredDisplayAttributes.includes(attr))
            .sort(([attrA], [attrB]) => filteredDisplayAttributes.indexOf(attrA) - filteredDisplayAttributes.indexOf(attrB)),
        );

        // get types

        const types = schema.nodes.find((n: any) => {
          let labelNode = '';
          if (node.label !== undefined) {
            labelNode = node.label;
          } else {
            const idParts = node._id.split('/');
            labelNode = idParts[0];
          }

          return labelNode === n.name;
        })?.attributes;

        if (types) {
          return {
            attribute: filteredAttributes,
            type: Object.fromEntries(types.map((t: any) => [t.name, t.type])),
          };
        } else {
          return {
            attribute: filteredAttributes,
            type: {},
          };
        }
      });
  }, [data.nodes, configuration.displayEntity, configuration.displayAttributes]);

  return (
    <div className="h-full w-full" ref={ref}>
      {attributesArray.length > 0 && (
        <Table
          data={attributesArray}
          itemsPerPage={configuration.itemsPerPage}
          showBarPlot={configuration.showBarplot}
          showAttributes={savedDisplayAttributes}
          selectedEntity={configuration.displayEntity}
        />
      )}
    </div>
  );
};

const TableSettings = ({
  configuration,
  graphMetadata,
  updateSettings,
}: {
  configuration: TableProps;
  graphMetadata: GraphMetadata;
  updateSettings: (val: any) => void;
}) => {
  useEffect(() => {
    if (graphMetadata && graphMetadata.nodes && graphMetadata.nodes.labels.length > 0) {
      updateSettings({ displayEntity: graphMetadata.nodes.labels[0] });
    }
  }, [graphMetadata]);

  const selectedNodeAttributes = useMemo(() => {
    if (configuration.displayEntity) {
      const nodeType = graphMetadata.nodes.types[configuration.displayEntity];
      if (nodeType && nodeType.attributes) {
        return Object.keys(nodeType.attributes);
      }
    }
    return [];
  }, [configuration.displayEntity, graphMetadata]);

  useEffect(() => {
    if (graphMetadata && graphMetadata.nodes && graphMetadata.nodes.labels.length > 0) {
      updateSettings({ displayAttributes: selectedNodeAttributes });
    }
  }, [selectedNodeAttributes]);

  return (
    <SettingsContainer>
      <Input
        type="dropdown"
        label="Select entity"
        value={configuration.displayEntity}
        onChange={(val) => updateSettings({ displayEntity: val })}
        options={graphMetadata.nodes.labels}
      />
      <Input
        type="boolean"
        label="Show barplot"
        value={configuration.showBarplot}
        onChange={(val) => updateSettings({ showBarplot: val })}
      />
      <Input
        type="dropdown"
        label="Items per page"
        value={configuration.itemsPerPage}
        onChange={(val) => updateSettings({ itemsPerPage: val })}
        options={[10, 25, 50, 100]}
      />
      <div>
        <span className="text-sm">Attributes to display</span>
        <div className="">
          <Input
            type="checkbox"
            value={configuration.displayAttributes}
            options={selectedNodeAttributes}
            onChange={(val: string[] | string) => {
              const updatedVal = Array.isArray(val) ? val : [val];
              updateSettings({ displayAttributes: updatedVal });
            }}
          />
        </div>
      </div>
    </SettingsContainer>
  );
};

export const TableComponent: VISComponentType = {
  displayName: 'TableVis',
  component: TableVis,
  settings: TableSettings,
  configuration: configuration,
};

export default TableComponent;
